import binascii
import datetime
import struct
import time
import zlib
from typing import Any, Optional, Type, Union

from cryptography.hazmat.primitives.hmac import HMAC
from django.conf import settings
from django.core.signing import (
    BadSignature,
    JSONSerializer,
    SignatureExpired,
    b64_decode,
    b64_encode,
    get_cookie_signer,
)
from django.utils.encoding import force_bytes
from django.utils.regex_helper import _lazy_re_compile

from ..typing import Algorithm, Serializer
from ..utils.crypto import HASHES, InvalidAlgorithm, constant_time_compare, salted_hmac

try:
    from django.core.signing import b62_decode, b62_encode  # type: ignore
except ImportError:
    from django.utils import baseconv

    # Required for Django 3.2 support
    b62_decode, b62_encode = baseconv.base62.decode, baseconv.base62.encode

__all__ = [
    "BadSignature",
    "SignatureExpired",
    "b64_encode",
    "b64_decode",
    "base64_hmac",
    "get_cookie_signer",
    "JSONSerializer",
    "dumps",
    "loads",
    "Signer",
    "TimestampSigner",
    "BytesSigner",
    "FernetSigner",
]

_MAX_CLOCK_SKEW = 60
_SEP_UNSAFE = _lazy_re_compile(r"^[A-z0-9-_=]*$")


def base64_hmac(
    salt: str,
    value: Union[bytes, str],
    key: Union[bytes, str],
    algorithm: Algorithm = "sha1",
) -> str:
    return b64_encode(
        salted_hmac(salt, value, key, algorithm=algorithm).finalize()
    ).decode()


def dumps(
    obj: Any,
    key: Optional[Union[bytes, str]] = None,
    salt: str = "django.core.signing",
    serializer: Type[Serializer] = JSONSerializer,
    compress: bool = False,
) -> str:
    """
    Return URL-safe, hmac signed base64 compressed JSON string. If key is
    None, use settings.SECRET_KEY instead. The hmac algorithm is the default
    Signer algorithm.

    If compress is True (not the default), check if compressing using zlib can
    save some space. Prepend a '.' to signify compression. This is included
    in the signature, to protect against zip bombs.

    Salt can be used to namespace the hash, so that a signed string is
    only valid for a given namespace. Leaving this at the default
    value or re-using a salt value across different parts of your
    application without good cause is a security risk.

    The serializer is expected to return a bytestring.
    """
    return TimestampSigner(key, salt=salt).sign_object(
        obj, serializer=serializer, compress=compress
    )


def loads(
    s: str,
    key: Optional[Union[bytes, str]] = None,
    salt: str = "django.core.signing",
    serializer: Type[Serializer] = JSONSerializer,
    max_age: Optional[Union[int, datetime.timedelta]] = None,
) -> Any:
    """
    Reverse of dumps(), raise BadSignature if signature fails.

    The serializer is expected to accept a bytestring.
    """
    return TimestampSigner(key, salt=salt).unsign_object(
        s, serializer=serializer, max_age=max_age
    )


class Signer:
    def __init__(
        self,
        key: Optional[Union[bytes, str]] = None,
        sep: str = ":",
        salt: Optional[str] = None,
        algorithm: Optional[Algorithm] = None,
    ) -> None:
        # Use of native strings in all versions of Python
        self.key = key or settings.SECRET_KEY
        self.sep = sep
        if _SEP_UNSAFE.match(self.sep):
            raise ValueError(
                "Unsafe Signer separator: %r (cannot be empty or consist of "
                "only A-z0-9-_=)" % sep,
            )
        self.salt = salt or f"{self.__class__.__module__}.{self.__class__.__name__}"
        self.algorithm = algorithm or "sha256"

    def signature(self, value: Union[bytes, str]) -> str:
        return base64_hmac(
            self.salt + "signer", value, self.key, algorithm=self.algorithm
        )

    def sign(self, value: str) -> str:
        return f"{value}{self.sep}{self.signature(value)}"

    def unsign(self, signed_value: str) -> str:
        if self.sep not in signed_value:
            raise BadSignature('No "%s" found in value' % self.sep)
        value, sig = signed_value.rsplit(self.sep, 1)
        if constant_time_compare(sig, self.signature(value)):
            return value
        raise BadSignature('Signature "%s" does not match' % sig)

    def sign_object(
        self,
        obj: Any,
        serializer: Type[Serializer] = JSONSerializer,
        compress: bool = False,
    ) -> str:
        """
        Return URL-safe, hmac signed base64 compressed JSON string.

        If compress is True (not the default), check if compressing using zlib
        can save some space. Prepend a '.' to signify compression. This is
        included in the signature, to protect against zip bombs.

        The serializer is expected to return a bytestring.
        """
        data = serializer().dumps(obj)
        # Flag for if it's been compressed or not.
        is_compressed = False

        if compress:
            # Avoid zlib dependency unless compress is being used.
            compressed = zlib.compress(data)
            if len(compressed) < (len(data) - 1):
                data = compressed
                is_compressed = True
        base64d = b64_encode(data).decode()
        if is_compressed:
            base64d = "." + base64d
        return self.sign(base64d)

    def unsign_object(
        self,
        signed_obj: str,
        serializer: Type[Serializer] = JSONSerializer,
        **kwargs: Any,
    ) -> Any:
        # Signer.unsign() returns str but base64 and zlib compression operate
        # on bytes.
        base64d = self.unsign(signed_obj, **kwargs).encode()
        decompress = base64d[:1] == b"."
        if decompress:
            # It's compressed; uncompress it first.
            base64d = base64d[1:]
        data = b64_decode(base64d)
        if decompress:
            data = zlib.decompress(data)
        return serializer().loads(data)


class TimestampSigner(Signer):
    def timestamp(self) -> str:
        return b62_encode(int(time.time()))

    def sign(self, value: str) -> str:
        value = f"{value}{self.sep}{self.timestamp()}"
        return super().sign(value)

    def unsign(
        self,
        value: str,
        max_age: Optional[Union[int, float, datetime.timedelta]] = None,
    ) -> str:
        """
        Retrieve original value and check it wasn't signed more
        than max_age seconds ago.
        """
        result = super().unsign(value)
        value, timestamp = result.rsplit(self.sep, 1)
        if max_age is not None:
            if isinstance(max_age, datetime.timedelta):
                max_age = max_age.total_seconds()
            # Check timestamp is not older than max_age
            age = time.time() - b62_decode(timestamp)
            if age > max_age:
                raise SignatureExpired(f"Signature age {age} > {max_age} seconds")
        return value


class BytesSigner:
    def __init__(
        self,
        key: Optional[Union[bytes, str]] = None,
        salt: Optional[str] = None,
        algorithm: Optional[Algorithm] = None,
    ) -> None:
        self.key = key or settings.SECRET_KEY
        self.salt = salt or f"{self.__class__.__module__}.{self.__class__.__name__}"
        self.algorithm = algorithm or "sha256"

        try:
            hasher = HASHES[self.algorithm]
        except KeyError as e:
            raise InvalidAlgorithm(
                "%r is not an algorithm accepted by the cryptography module."
                % algorithm
            ) from e

        self._digest_size = hasher.digest_size

    def signature(self, value: Union[bytes, str]) -> bytes:
        return salted_hmac(
            self.salt + "signer", value, self.key, algorithm=self.algorithm
        ).finalize()

    def sign(self, value: Union[bytes, str]) -> bytes:
        return force_bytes(value) + self.signature(value)

    def unsign(self, signed_value: bytes) -> bytes:
        value, sig = (
            signed_value[: -self._digest_size],
            signed_value[-self._digest_size :],
        )
        if constant_time_compare(sig, self.signature(value)):
            return value
        raise BadSignature('Signature "%r" does not match' % binascii.b2a_base64(sig))


class FernetSigner:
    version = b"\x80"

    def __init__(
        self,
        key: Optional[Union[bytes, str]] = None,
        algorithm: Optional[Algorithm] = None,
    ) -> None:
        self.key = key or settings.SECRET_KEY
        self.algorithm = algorithm or "sha256"

        try:
            hasher = HASHES[self.algorithm]
        except KeyError as e:
            raise InvalidAlgorithm(
                "%r is not an algorithm accepted by the cryptography module."
                % algorithm
            ) from e

        self.hasher = hasher

    def signature(self, value: Union[bytes, str]) -> bytes:
        h = HMAC(
            force_bytes(self.key),
            self.hasher,
            backend=settings.CRYPTOGRAPHY_BACKEND,  # type: ignore
        )
        h.update(force_bytes(value))
        return h.finalize()

    def sign(self, value: Union[bytes, str], current_time: int) -> bytes:
        payload = struct.pack(">cQ", self.version, current_time)
        payload += force_bytes(value)
        return payload + self.signature(payload)

    def unsign(
        self,
        signed_value: bytes,
        max_age: Optional[Union[int, float, datetime.timedelta]] = None,
    ) -> bytes:
        """
        Retrieve original value and check it wasn't signed more
        than max_age seconds ago.
        """
        h_size, d_size = struct.calcsize(">cQ"), self.hasher.digest_size
        fmt = ">cQ%ds%ds" % (len(signed_value) - h_size - d_size, d_size)
        try:
            version, timestamp, value, sig = struct.unpack(fmt, signed_value)
        except struct.error as err:
            raise BadSignature("Signature is not valid") from err
        if version != self.version:
            raise BadSignature("Signature version not supported")
        if max_age is not None:
            if isinstance(max_age, datetime.timedelta):
                max_age = max_age.total_seconds()
            # Check timestamp is not older than max_age
            age = abs(time.time() - timestamp)
            if age > max_age + _MAX_CLOCK_SKEW:
                raise SignatureExpired(f"Signature age {age} > {max_age} seconds")
        if constant_time_compare(sig, self.signature(signed_value[:-d_size])):
            return value
        raise BadSignature('Signature "%r" does not match' % binascii.b2a_base64(sig))



from django.db import migrations, models

import django_cryptography.fields


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="EncryptedCharModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", models.CharField(max_length=15)),
            ],
        ),
        migrations.CreateModel(
            name="EncryptedDateTimeModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("datetime", models.DateTimeField()),
                ("date", models.DateField()),
                ("time", models.TimeField()),
                ("auto_now", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="EncryptedIntegerModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="EncryptedNullableIntegerModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", models.IntegerField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name="EncryptedTTLIntegerModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="DefaultPickledModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "field",
                    django_cryptography.fields.PickledField(default=b""),
                ),
            ],
        ),
        migrations.CreateModel(
            name="NullablePickledModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "field",
                    django_cryptography.fields.PickledField(blank=True, null=True),
                ),
            ],
        ),
        migrations.CreateModel(
            name="OtherEncryptedTypesModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("ip", models.GenericIPAddressField()),
                ("uuid", models.UUIDField()),
                ("decimal", models.DecimalField(decimal_places=2, max_digits=5)),
            ],
        ),
        migrations.CreateModel(
            name="PickledModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", django_cryptography.fields.PickledField()),
            ],
        ),
    ]



import datetime
from typing import Any, Optional, Union

from typing_extensions import Literal, Protocol

Algorithm = Literal[
    "blake2b",
    "blake2s",
    "md5",
    "sha1",
    "sha224",
    "sha256",
    "sha384",
    "sha3_224",
    "sha3_256",
    "sha3_384",
    "sha3_512",
    "sha512",
    "sha512_224",
    "sha512_256",
    "sm3",
]


class DBAPI(Protocol):
    def Binary(self, obj: Union[bytes, str]) -> Any:
        ...


class DatabaseWrapper(Protocol):
    Database: DBAPI


class Serializer(Protocol):
    def dumps(self, obj: Any) -> bytes:
        ...

    def loads(self, data: bytes) -> Any:
        ...


class Signer(Protocol):
    def __init__(
        self, key: Optional[Union[bytes, str]] = None, algorithm: Optional[str] = None
    ) -> None:
        ...

    def signature(self, value: Union[bytes, str]) -> bytes:
        ...

    def sign(self, value: Union[bytes, str], current_time: int) -> bytes:
        ...

    def unsign(
        self,
        signed_value: bytes,
        max_age: Optional[Union[int, float, datetime.timedelta]] = None,
    ) -> bytes:
        ...



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("fields", "0001_initial"),
    ]

    operations = [
        migrations.RenameField(
            model_name="encryptedcharmodel",
            old_name="field",
            new_name="old_field",
        ),
        migrations.RenameField(
            model_name="encrypteddatetimemodel",
            old_name="auto_now",
            new_name="old_auto_now",
        ),
        migrations.RenameField(
            model_name="encrypteddatetimemodel",
            old_name="date",
            new_name="old_date",
        ),
        migrations.RenameField(
            model_name="encrypteddatetimemodel",
            old_name="datetime",
            new_name="old_datetime",
        ),
        migrations.RenameField(
            model_name="encrypteddatetimemodel",
            old_name="time",
            new_name="old_time",
        ),
        migrations.RenameField(
            model_name="encryptedintegermodel",
            old_name="field",
            new_name="old_field",
        ),
        migrations.RenameField(
            model_name="encryptednullableintegermodel",
            old_name="field",
            new_name="old_field",
        ),
        migrations.RenameField(
            model_name="encryptedttlintegermodel",
            old_name="field",
            new_name="old_field",
        ),
        migrations.RenameField(
            model_name="otherencryptedtypesmodel",
            old_name="decimal",
            new_name="old_decimal",
        ),
        migrations.RenameField(
            model_name="otherencryptedtypesmodel",
            old_name="ip",
            new_name="old_ip",
        ),
        migrations.RenameField(
            model_name="otherencryptedtypesmodel",
            old_name="uuid",
            new_name="old_uuid",
        ),
    ]



from django.db import models

from django_cryptography.fields import PickledField, encrypt


class PickledModel(models.Model):
    field = PickledField()


class DefaultPickledModel(models.Model):
    field = PickledField(default=b"")


class NullablePickledModel(models.Model):
    field = PickledField(blank=True, null=True)


class EncryptedIntegerModel(models.Model):
    field = encrypt(models.IntegerField())


class EncryptedNullableIntegerModel(models.Model):
    field = encrypt(models.IntegerField(blank=True, null=True))


class EncryptedTTLIntegerModel(models.Model):
    field = encrypt(models.IntegerField(), ttl=60)


class EncryptedCharModel(models.Model):
    field = encrypt(models.CharField(max_length=15))


class EncryptedDateTimeModel(models.Model):
    datetime = encrypt(models.DateTimeField())
    date = encrypt(models.DateField())
    time = encrypt(models.TimeField())
    auto_now = encrypt(models.DateTimeField(auto_now=True))


class OtherEncryptedTypesModel(models.Model):
    ip = encrypt(models.GenericIPAddressField())
    uuid = encrypt(models.UUIDField())
    decimal = encrypt(models.DecimalField(max_digits=5, decimal_places=2))


@encrypt
class EncryptedFieldSubclass(models.IntegerField):
    pass



import json
import pickle

from django.core import exceptions, serializers
from django.db import IntegrityError
from django.test import TestCase
from django.utils import timezone

from django_cryptography.fields import PickledField

from .models import DefaultPickledModel, NullablePickledModel, PickledModel


class TestSaveLoad(TestCase):
    def test_integer(self):
        instance = PickledModel(field=42)
        instance.save()
        loaded = PickledModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_string(self):
        instance = PickledModel(field="Hello, world!")
        instance.save()
        loaded = PickledModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_datetime(self):
        instance = PickledModel(field=timezone.now())
        instance.save()
        loaded = PickledModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_default(self):
        instance = DefaultPickledModel()
        instance.save()
        loaded = DefaultPickledModel.objects.get(pk=instance.pk)
        self.assertEqual(loaded.field, b"")
        self.assertEqual(instance.field, loaded.field)

    def test_default_null(self):
        instance = NullablePickledModel()
        instance.save()
        loaded = NullablePickledModel.objects.get(pk=instance.pk)
        self.assertEqual(loaded.field, None)
        self.assertEqual(instance.field, loaded.field)

    def test_null_handling(self):
        instance = NullablePickledModel(field=None)
        instance.save()
        loaded = NullablePickledModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

        instance = PickledModel(field=None)
        with self.assertRaises(IntegrityError):
            instance.save()


class TestQuerying(TestCase):
    def setUp(self):
        self.objs = [
            NullablePickledModel.objects.create(field=[1]),
            NullablePickledModel.objects.create(field=[2]),
            NullablePickledModel.objects.create(field=[2, 3]),
            NullablePickledModel.objects.create(field=[20, 30, 40]),
            NullablePickledModel.objects.create(field=None),
        ]

    def test_exact(self):
        self.assertSequenceEqual(
            NullablePickledModel.objects.filter(field__exact=[1]), self.objs[:1]
        )

    def test_isnull(self):
        self.assertSequenceEqual(
            NullablePickledModel.objects.filter(field__isnull=True), self.objs[-1:]
        )

    def test_in(self):
        self.assertSequenceEqual(
            NullablePickledModel.objects.filter(field__in=[[1], [2]]), self.objs[:2]
        )

    def test_unsupported(self):
        with self.assertRaises(exceptions.FieldError):
            NullablePickledModel.objects.filter(field__contains=[2])


class TestMigrations(TestCase):
    def test_deconstruct(self):
        field = PickledField()
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual("django_cryptography.fields.PickledField", path)
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {})


class TestSerialization(TestCase):
    test_data = (
        (
            # Python 3.4
            '[{"fields": {"field": "gANdcQAoSwFLAk5lLg=="}, '
            '"model": "fields.pickledmodel", "pk": null}]'
        )
        if pickle.HIGHEST_PROTOCOL < 5
        else (
            # Python 3.8
            '[{"fields": {"field": "gASVCgAAAAAAAABdlChLAUsCTmUu"}, '
            '"model": "fields.pickledmodel", "pk": null}]'
        )
    )

    def test_dumping(self):
        instance = PickledModel(field=[1, 2, None])
        data = serializers.serialize("json", [instance])
        self.assertEqual(json.loads(self.test_data), json.loads(data))

    def test_loading(self):
        instance = list(serializers.deserialize("json", self.test_data))[0].object
        self.assertEqual([1, 2, None], instance.field)


class TestValidation(TestCase):
    def test_validate(self):
        field = PickledField()
        field.clean(None, None)



DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
    }
}

SECRET_KEY = "django_tests_secret_key"

INSTALLED_APPS = [
    "tests.fields",
]

SIGNING_BACKEND = "django_cryptography.core.signing.TimestampSigner"

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"



import base64
import os
import time
from binascii import Error
from typing import Dict, Optional, Union

from cryptography.hazmat.primitives import constant_time, hashes, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.hmac import HMAC
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from django.utils import crypto
from django.utils.encoding import force_bytes

from ..conf import CryptographyConf
from ..typing import Algorithm, Signer

settings = CryptographyConf()


# AddedInDjango30Warning: Remove when Django 2.2 leaves LTS
class InvalidAlgorithm(ValueError):
    """Algorithm is not supported by cryptography."""

    pass


class InvalidToken(Exception):
    pass


HASHES: Dict[Algorithm, hashes.HashAlgorithm] = {
    "blake2b": hashes.BLAKE2b(64),
    "blake2s": hashes.BLAKE2s(32),
    "md5": hashes.MD5(),
    "sha1": hashes.SHA1(),
    "sha224": hashes.SHA224(),
    "sha256": hashes.SHA256(),
    "sha384": hashes.SHA384(),
    "sha3_224": hashes.SHA3_224(),
    "sha3_256": hashes.SHA3_256(),
    "sha3_384": hashes.SHA3_384(),
    "sha3_512": hashes.SHA3_512(),
    "sha512": hashes.SHA512(),
    "sha512_224": hashes.SHA512_224(),
    "sha512_256": hashes.SHA512_256(),
    "sm3": hashes.SM3(),
}


def salted_hmac(
    key_salt: Union[bytes, str],
    value: Union[bytes, str],
    secret: Optional[Union[bytes, str]] = None,
    *,
    algorithm: Algorithm = "sha1",
) -> HMAC:
    """
    Return the HMAC of 'value', using a key generated from key_salt and a
    secret (which defaults to settings.SECRET_KEY). Default algorithm is SHA1,
    but any algorithm name supported by cryptography can be passed.

    A different key_salt should be passed in for every application of HMAC.
    """
    if secret is None:
        secret = settings.SECRET_KEY

    key_salt = force_bytes(key_salt)
    secret = force_bytes(secret)
    try:
        hasher = HASHES[algorithm]
    except KeyError as e:
        raise InvalidAlgorithm(
            "%r is not an algorithm accepted by the cryptography module." % algorithm
        ) from e

    # We need to generate a derived key from our base key.  We can do this by
    # passing the key_salt and our base key through a pseudo-random function.
    digest = hashes.Hash(hasher, backend=settings.CRYPTOGRAPHY_BACKEND)
    digest.update(key_salt + secret)
    key = digest.finalize()

    # If len(key_salt + secret) > sha_constructor().block_size, the above
    # line is redundant and could be replaced by key = key_salt + secret, since
    # the hmac module does the same thing for keys longer than the block size.
    # However, we need to ensure that we *always* do this.
    h = HMAC(key, hasher, backend=settings.CRYPTOGRAPHY_BACKEND)
    h.update(force_bytes(value))
    return h


get_random_string = crypto.get_random_string


def constant_time_compare(val1: Union[bytes, str], val2: Union[bytes, str]) -> bool:
    """Return True if the two strings are equal, False otherwise."""
    return constant_time.bytes_eq(force_bytes(val1), force_bytes(val2))


def pbkdf2(
    password: Union[bytes, str],
    salt: Union[bytes, str],
    iterations: int,
    dklen: int = 0,
    digest: Optional[hashes.HashAlgorithm] = None,
) -> bytes:
    """
    Implements PBKDF2 with the same API as Django's existing
    implementation, using cryptography.
    """
    if digest is None:
        digest = hashes.SHA256()
    dklen = dklen or digest.digest_size
    password = force_bytes(password)
    salt = force_bytes(salt)
    kdf = PBKDF2HMAC(
        digest, dklen, salt, iterations, backend=settings.CRYPTOGRAPHY_BACKEND
    )
    return kdf.derive(password)


class FernetBytes:
    """
    This is a modified version of the Fernet encryption algorithm from
    the Python Cryptography library. The main change is the allowance
    of varied length cryptographic keys from the base 128-bit. There is
    also an emphasis on using Django's settings system for sane defaults.
    """

    def __init__(
        self, key: Optional[Union[bytes, str]] = None, signer: Optional[Signer] = None
    ) -> None:
        if signer is None:
            from ..core.signing import FernetSigner

            signer = FernetSigner()
        self.key = key or settings.CRYPTOGRAPHY_KEY
        self.signer = signer

    def encrypt(self, data: Union[bytes, str]) -> bytes:
        return self.encrypt_at_time(data, int(time.time()))

    def encrypt_at_time(self, data: Union[bytes, str], current_time: int) -> bytes:
        data = force_bytes(data)
        iv = os.urandom(16)
        return self._encrypt_from_parts(data, current_time, iv)

    def _encrypt_from_parts(self, data: bytes, current_time: int, iv: bytes) -> bytes:
        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        padded_data = padder.update(data) + padder.finalize()
        encryptor = Cipher(
            algorithms.AES(force_bytes(self.key)),
            modes.CBC(iv),
            backend=settings.CRYPTOGRAPHY_BACKEND,
        ).encryptor()
        ciphertext = encryptor.update(padded_data) + encryptor.finalize()

        return self.signer.sign(iv + ciphertext, current_time)

    def decrypt(self, data: bytes, ttl: Optional[int] = None) -> bytes:
        data = self.signer.unsign(data, ttl)

        iv = data[:16]
        ciphertext = data[16:]
        decryptor = Cipher(
            algorithms.AES(force_bytes(self.key)),
            modes.CBC(iv),
            backend=settings.CRYPTOGRAPHY_BACKEND,
        ).decryptor()
        plaintext_padded = decryptor.update(ciphertext)
        try:
            plaintext_padded += decryptor.finalize()
        except ValueError as err:
            raise InvalidToken from err

        # Remove padding
        unpadder = padding.PKCS7(algorithms.AES.block_size).unpadder()
        unpadded = unpadder.update(plaintext_padded)
        try:
            unpadded += unpadder.finalize()
        except ValueError as err:
            raise InvalidToken from err
        return unpadded


class Fernet(FernetBytes):
    def __init__(
        self, key: Optional[Union[bytes, str]] = None, signer: Optional[Signer] = None
    ) -> None:
        if signer is None:
            from ..core.signing import FernetSigner

            signer = FernetSigner()
        if key is None:
            super().__init__()
        else:
            key = base64.urlsafe_b64decode(key)
            if len(key) != 32:
                raise ValueError("Fernet key must be 32 url-safe base64-encoded bytes.")

            super().__init__(key[16:], type(signer)(key[:16]))

    def _encrypt_from_parts(self, data: bytes, current_time: int, iv: bytes) -> bytes:
        payload = super()._encrypt_from_parts(data, current_time, iv)
        return base64.urlsafe_b64encode(payload)

    def decrypt(self, token: bytes, ttl: Optional[int] = None) -> bytes:
        try:
            data = base64.urlsafe_b64decode(token)
        except (TypeError, Error) as err:
            raise InvalidToken from err
        return super().decrypt(data, ttl)



from django.db import migrations, models

import django_cryptography.fields


class Migration(migrations.Migration):
    dependencies = [
        ("fields", "0002_rename_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="encryptedcharmodel",
            name="field",
            field=django_cryptography.fields.encrypt(
                models.CharField(default=None, max_length=15)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="encrypteddatetimemodel",
            name="auto_now",
            field=django_cryptography.fields.encrypt(
                models.DateTimeField(default=None, auto_now=True)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="encrypteddatetimemodel",
            name="date",
            field=django_cryptography.fields.encrypt(models.DateField(default=None)),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="encrypteddatetimemodel",
            name="datetime",
            field=django_cryptography.fields.encrypt(
                models.DateTimeField(default=None)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="encrypteddatetimemodel",
            name="time",
            field=django_cryptography.fields.encrypt(models.TimeField(default=None)),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="encryptedintegermodel",
            name="field",
            field=django_cryptography.fields.encrypt(models.IntegerField(default=None)),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="encryptednullableintegermodel",
            name="field",
            field=django_cryptography.fields.encrypt(
                models.IntegerField(default=None, blank=True, null=True)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="encryptedttlintegermodel",
            name="field",
            field=django_cryptography.fields.encrypt(
                models.IntegerField(default=None), ttl=60
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="otherencryptedtypesmodel",
            name="decimal",
            field=django_cryptography.fields.encrypt(
                models.DecimalField(default=None, decimal_places=2, max_digits=5)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="otherencryptedtypesmodel",
            name="ip",
            field=django_cryptography.fields.encrypt(
                models.GenericIPAddressField(default=None)
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="otherencryptedtypesmodel",
            name="uuid",
            field=django_cryptography.fields.encrypt(models.UUIDField(default=None)),
            preserve_default=False,
        ),
    ]



from django.db import migrations, models

import django_cryptography.fields


class Migration(migrations.Migration):
    dependencies = [
        ("fields", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="integerencrypteddefaultmodel",
            name="field_2",
            field=django_cryptography.fields.encrypt(
                models.IntegerField(max_length=50, blank=True)
            ),
            preserve_default=False,
        ),
    ]
